import bpy
from re import split
from bpy.props import (FloatProperty, IntProperty, BoolProperty, StringProperty, CollectionProperty, PointerProperty)
from bpy.types import (Operator, PropertyGroup, World, AddonPreferences, RENDER_PT_format, Panel)
from bpy.app.handlers import persistent  

# ======================================================
#                       GLOBALS
# ======================================================
camera_render_props_list = [
    # you can add property for your own there
    #    Property Name                #          Property Defferer                      
    # Format
    ('render.resolution_x',                IntProperty(default=1024)                   ),               
    ('render.resolution_y',                IntProperty(default=1024)                   ),               
    ('render.resolution_percentage',       IntProperty(default=100)                    ),
    ('render.pixel_aspect_x',              FloatProperty(default=1)                    ),                                     
    ('render.pixel_aspect_y',              FloatProperty(default=1)                    ),                                     
    # 
    ('render.use_border',                  BoolProperty (default=1)                    ),                                            
    ('render.use_crop_to_border',          BoolProperty (default=1)                    ),                                            
    ('render.fps',                         IntProperty  (default=24)                   ),                         
    ('render.fps_base',                    FloatProperty(default=1.000)                ),            
    # Color Management
    ('view_settings.view_transform',       StringProperty(default='Filmic')            ), 
    ('view_settings.look',                 StringProperty(default='None')              ), 
    ('view_settings.exposure',             FloatProperty (default=0)                   ), 
    ('view_settings.gamma',                FloatProperty (default=1)                   ), 
    # Sampling
    ('cycles.use_adaptive_sampling',       BoolProperty  (default=0)                   ), 
    ('cycles.samples',                     IntProperty   (default=4096)                ), 
    ('cycles.time_limit',                  FloatProperty (default=60)                  ), 
    ('cycles.use_denoising',               BoolProperty  (default=1)                   ), 
    ('cycles.denoiser',                    StringProperty(default='OPTIX')             ), 
    ('cycles.denoising_input_passes',      StringProperty(default='RGB_ALBEDO_NORMAL') ), 
    # Light Paths
    ('cycles.max_bounces',                 IntProperty   (default=8)                   ), 
    ('cycles.diffuse_bounces',             IntProperty   (default=8)                   ), 
    ('cycles.glossy_bounces',              IntProperty   (default=8)                   ), 
    ('cycles.transmission_bounces',        IntProperty   (default=8)                   ), 
    ('cycles.volume_bounces',              IntProperty   (default=0)                   ), 
    ('cycles.transparent_max_bounces',     IntProperty   (default=12)                  ), 
    # Light Paths: Caustincs
    ('cycles.blur_glossy',                 FloatProperty (default=1)                   ), 
    ('cycles.caustics_reflective',         BoolProperty  (default=0)                   ), 
    ('cycles.caustics_refractive',         BoolProperty  (default=0)                   ), 
    # Light Paths: Fast GI Approx
    ('cycles.use_fast_gi',                 BoolProperty  (default=0)                   ), 
    ('cycles.fast_gi_method',              StringProperty(default='REPLACE')           ), 
    ('cycles.ao_bounces',                  IntProperty   (default=1)                   ), 
    ('cycles.ao_bounces_render',           FloatProperty (default=1)                   ), 
    # Simplify
    ('render.use_simplify',                BoolProperty  (default=0)                   ), 
    ('render.simplify_subdivision',        IntProperty   (default=12)                  ), 
    ('render.simplify_child_particles',    FloatProperty (default=1)                   ), 
    ('cycles.texture_limit',               StringProperty(default='2048')              ), 
    ('render.simplify_volumes',            FloatProperty (default=1)                   ), 
    ('render.simplify_subdivision_render', IntProperty   (default=12)                  ), 
    ('render.simplify_subdivision_render', IntProperty   (default=12)                  ), 
    ('cycles.texture_limit_render',        StringProperty(default='4096')              ), 
    # Motion Blur
    ('render.use_motion_blur',             BoolProperty  (default=0)                   ), 
    ('render.motion_blur_position',        StringProperty(default='CENTER')            ), 
    ('render.motion_blur_shutter',         FloatProperty (default=.5)                  ), 
    # stray properties
    ('world',                            PointerProperty(type = World, name = 'world'), ),
    ]

#                         [       i[0]           ,        i[1]           , [ i[2][0], i[2][1] ]  ,    i[3]         ]
#                         [ 'cycles.samples'     ,  'cycles__samples'    , ['cycles','samples']  , IntProperty(...)]
camera_render_props_zip = [ (     i[0]           , i[0].replace('.','__'),  split(r'\.', i[0])   ,    i[1]         ) for i in camera_render_props_list ]

# ======================================================
#                       FUCNS
# ======================================================
def check_camera_type(func):
    '''@ decorator: pass function if there is no active camera'''
    def wrapper(*args):
        if bpy.context.scene.camera is not None:
            if bpy.context.scene.camera.type == 'CAMERA':
#                print('check_camera_type OK')
                return func(*args)
        else: pass
    return wrapper

def addon_preferences():
    return bpy.context.preferences.addons[__package__].preferences
# ======================================================
#                  MSGBUS registering
# ======================================================
@check_camera_type
def store_props_in_object_data(x,y, prnt):
    '''lambda func for msgbus'''
#    print(y, prnt)
    setattr(bpy.context.scene.camera.data.Individual_Camera_Properties_CAMERA_RENDER_PG, x, bpy.context.scene.path_resolve(y))
    return None

Individual_Camera_Properties_msgbus_owner = object()

@persistent
def Individual_Camera_Properties_msgbus(scene, context):
    '''used in bpy.app.handlers.load_post'''
    #  It would be more relevant to use `type` `bpy_struct` as the key in subscription,
    #  but since `bpy.types.CyclesRenderSettings` has no real RNA structures (see issue #68050)
    #  we're using context data as key
    bpy.msgbus.clear_by_owner(Individual_Camera_Properties_msgbus_owner)
    preferences = addon_preferences()
    for i in camera_render_props_zip:
        '''dynamically subscribe properties from the list'''
        if preferences.path_resolve(i[1]): # Check if property enabled in preferences
            if i[0] == i[1]: key = (bpy.types.Scene, i[0]) # Using key subscribtion due to inability to use `path_resolve` on `bpy.context.scene.world`
            else           : key = bpy.context.scene.path_resolve(i[0], False)    
            bpy.msgbus.subscribe_rna(
            key = key,
            owner = Individual_Camera_Properties_msgbus_owner,
            args = (),
            notify = lambda x=i[1], y=i[0]: store_props_in_object_data(x,y,'prop was updated'),
            )
        
    print('Individual Camera Properties msgbus has been registered, unregistering previous if exist')
# ======================================================
#                Active Camera Handler
# ======================================================
@check_camera_type
def set_cam_render_props(camera_render_props_zip):
    addon_prefs = addon_preferences()
    '''set render scene parameters from props stored in CAMERA RENDER'''
    for i in camera_render_props_zip:
        if addon_prefs.path_resolve(i[1]): # AddonPreferences
            if i[0] == i[1]: # stray properties (for now 'world' prop only)
              setattr(bpy.context.scene, i[0], getattr(getattr(bpy.context.scene.camera.data, 'Individual_Camera_Properties_CAMERA_RENDER_PG'), i[0]))
            else:
              setattr(getattr(bpy.context.scene,  i[2][0]),        i[2][1], getattr(getattr(bpy.context.scene.camera.data, 'Individual_Camera_Properties_CAMERA_RENDER_PG'), i[1]))
    #         setattr(getattr(bpy.context.scene, 'render'), 'resolution_x', getattr(getattr(bpy.context.scene.camera,      'Individual_Camera_Properties_CAMERA_RENDER_PG'), 'render__resolution_x'))
@persistent
def Individual_Camera_Properties_cam_change(scene, context):
    '''bpy.app.handlers.depsgraph_update_post'''          
    if not "cam" in scene.keys():
#        print('Active camera was NOT changed')
        pass
    elif scene["cam"] != scene.camera: 
        set_cam_render_props(camera_render_props_zip)
#        print('Active camera was changed:', scene.camera)
        scene["cam"] = scene.camera

@persistent
def Individual_Camera_Properties_initialize(scene, context):
    '''bpy.app.handlers.load_post'''
    Individual_Camera_Properties_msgbus(scene, context)
    bpy.context.scene['cam'] = bpy.context.scene.camera # variable to compare for camera change detection handler
    if bpy.context.scene.camera is not None:
        set_cam_render_props(camera_render_props_zip)
     
# ======================================================
#                         UI
# ======================================================
class Individual_Camera_Properties_UI_PT(Panel):
    """Creates a Panel in the Object Properties window"""
    bl_label = "Individual Camera Properties"
    bl_idname = "OBJECT_PT_individual_camera_properties"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "data"
    bl_options = {'DEFAULT_CLOSED'}
    
    @classmethod
    def poll(cls, context):
        if context.active_object is None:
            return False
        return context.active_object.type == 'CAMERA'
    
    def draw(self, context):
        object = context.object
        layout = self.layout
        row = layout.row(); row.label(text="For Copying Usage", icon='INFO')
#        layout.enabled = False
        layout.use_property_decorate = False  # No animation.
        layout.use_property_split = False
        col = layout.column(align=True)
        box = layout.box()
        col = box.column(align=True)
        PG = object.data.Individual_Camera_Properties_CAMERA_RENDER_PG    
        for i in list(PG.__annotations__):
            col.prop(PG, i, text=i.replace('__','.'))

# ======================================================
#                      CLASSES
# ======================================================
class Individual_Camera_Properties_AddonPreferences(AddonPreferences):
    # this must match the addon name, use '__package__'
    # when defining this in a submodule of a python package.
    bl_idname = __package__
    
    __annotations__ = {i[1]   : BoolProperty (default = False) for i in camera_render_props_zip}
    
    def draw(self, context):
        layout = self.layout
        col = layout.column() 
        col.alert = True
        col.label(text="After enabling addon press Update button bellow:", icon='WARNING_LARGE')
        col.operator("ui.individual_camera_properties_msgbus")
        col = layout.column(align=True)
        col.label(text="Choose what properties should be synchronized: (and press Update button)", icon='TAG')
        for i in camera_render_props_zip:
            col.prop(self, i[1], text=i[0])

class Individual_Camera_Properties_msgbus_update_Operator(Operator):
    bl_idname = "ui.individual_camera_properties_msgbus"
    bl_label = "Update"
    bl_description = "Updates msgbus for the addon. It needs to be done every time you determine properties bellow"
    bl_options = {'REGISTER', 'INTERNAL'}

    def execute(scene, context):
        Individual_Camera_Properties_initialize(scene, context)
        return {'FINISHED'}

class Individual_Camera_Properties_CAMERA_RENDER_PG(PropertyGroup):
    ''' Dynamically create props for camera RENDER properties. 
    User will use default properties in Render tab to change them: 
        they will be written to camera data automatically by msgbus'''
    __annotations__ = {i[1]   : i[3] for i in camera_render_props_zip} # creating prop without "render." prefix

# ======================================================
#                      REGISTER
# ======================================================
classes = [
    # PG
    Individual_Camera_Properties_CAMERA_RENDER_PG,
    # UI Panels
    Individual_Camera_Properties_UI_PT,
    # Preferences
    Individual_Camera_Properties_AddonPreferences,
    Individual_Camera_Properties_msgbus_update_Operator,
    ]

def register(): 
    # classes
    for c in classes: bpy.utils.register_class(c)
    # props
    bpy.types.Camera.Individual_Camera_Properties_CAMERA_RENDER_PG = PointerProperty(type = Individual_Camera_Properties_CAMERA_RENDER_PG)
    # handler
    bpy.app.handlers.depsgraph_update_post.append(Individual_Camera_Properties_cam_change)
    bpy.app.handlers.load_post.append(Individual_Camera_Properties_initialize)

def unregister(): 
    # classes
    for cls in classes: bpy.utils.unregister_class(cls)
    # handler
    bpy.app.handlers.load_post.remove(Individual_Camera_Properties_initialize)
    bpy.app.handlers.depsgraph_update_post.remove(Individual_Camera_Properties_cam_change)
    # props
    del bpy.types.Camera.Individual_Camera_Properties_CAMERA_RENDER_PG
    # msgbus
    bpy.msgbus.clear_by_owner(Individual_Camera_Properties_msgbus_owner)

if __name__ == "__main__": register()